@@ -1,4 +1,4 @@
-use std::sync::Arc;
+use std::{sync::Arc, collections::HashSet};
 
 use crossbeam::queue::SegQueue;
 use nacos_naming_client:: {
@@ -10,6 +10,7 @@ use nacos_naming_client:: {
 use async_once::AsyncOnce;
 use lazy_static::lazy_static;
 use async_trait::async_trait;
+use tokio::sync::Mutex;
 use tower::discover::Change;
 use tonic::transport::{Endpoint, Channel};
 use log;
@@ -65,16 +66,14 @@ fn parse_bool_env(key: &str, default: bool) -> bool {
     }).unwrap_or(default)
 }
 struct ChangeListener {
+    cur_endpoints: Arc<Mutex<HashSet<String>>>,
     queue: Arc<SegQueue<Change<String, Endpoint>>>,
 }
 
-#[async_trait]
-impl ServiceChangeListener for ChangeListener {
-    async fn changed(&self, service_name: &str, hosts: Vec<Instance>) {
-        log::debug!("obtain service[{}] change from nacos: {:?}", service_name, hosts);
-        for mut instance in hosts {
-            let port = instance.metadata.remove("gRPC_port").unwrap_or(instance.port.to_string());
-            let endpoint = format!("{}://{}:{}", "http", instance.ip, port);
+impl ChangeListener {
+    fn gen_endpoints<'a>(&self, list: impl IntoIterator<Item = &'a String>) -> Vec<(String, Endpoint)> {
+        let mut ret = Vec::new();
+        for endpoint in list {
             let ep = Endpoint::from_shared(endpoint.clone());
             let ep = match ep {
                 Ok(endpoint) => endpoint,
@@ -83,33 +82,64 @@ impl ServiceChangeListener for ChangeListener {
                     continue
                 } 
             };
-            let change = if instance.healthy && instance.enabled && instance.weight > 0f64 {
-                Change::Insert(endpoint, ep)
-            } else {
-                Change::Remove(endpoint)
-            };
+            ret.push((endpoint.clone(), ep));
+        }
+        ret
+    }
+}
 
+#[async_trait]
+impl ServiceChangeListener for ChangeListener {
+    async fn changed(&self, service_name: &str, hosts: Vec<Instance>) {
+        log::debug!("obtain service[{}] change from nacos: {:?}", service_name, hosts);
+        let endpoints = hosts.into_iter()
+            .filter(|instance| instance.healthy && instance.enabled && instance.weight > 0f64)
+            .map(|mut instance| {
+                let port = instance.metadata.remove("gRPC_port").unwrap_or(instance.port.to_string());
+                format!("{}://{}:{}", "http", instance.ip, port)
+            })
+            .collect::<HashSet<_>>();
+        
+        let mut cur = self.cur_endpoints.lock().await;
+
+        // remove
+        let diffs = cur.difference(&endpoints).map(|x| x.clone()).collect::<Vec<_>>();
+        for remove_diff in diffs {
+            cur.remove(&remove_diff);
+            let change = Change::Remove(remove_diff.clone());
             self.queue.push(change);
-            if self.queue.len() > 10 {
+            if self.queue.len() > 256 {
+                self.queue.pop();
+            }
+        }
+        
+        // insert
+        for (key, endpoint) in self.gen_endpoints(endpoints.difference(&cur)) {
+            cur.insert(key.clone());
+            let change = Change::Insert(key, endpoint);
+            self.queue.push(change);
+            if self.queue.len() > 256 {
                 self.queue.pop();
             }
         }
     }
 }
 
-
-
 pub async fn gen_channel(service_name: &str) -> Channel {
     let nacos_client = NACOS_CLIENT.get().await;
     let (channel, rx) = Channel::balance_channel(1);
     let queue = Arc::new(SegQueue::new());
-    let listener = ChangeListener {queue: queue.clone()};
+    let listener = ChangeListener {cur_endpoints: Arc::new(Mutex::new(HashSet::new())), queue: queue.clone()};
+    let name_cloned = service_name.to_string();
     tokio::spawn(async move {
         loop {
             if let Some(change) = queue.pop() {
+                log::info!("service[{}] change: {:?}", name_cloned, change);
                 if let Err(error) = rx.try_send(change) {
                     log::debug!("failed to modify endpoint list: {}", error)
                 }
+            } else {
+                tokio::time::sleep(std::time::Duration::from_secs(1)).await;
             }
         }
     });
